#pragma once

// TODO: figure out values

//
// Base offsets
//
#define STARTPCSOUNDS 0
#define STARTADLIBSOUNDS 43
#define STARTDIGISOUNDS 86
#define STARTMUSIC 119

#define LASTSOUND 43

#define INTRO_MUS 0
#define ALLGOOD_MUS 1
#define SONG3_MUS 2
#define SONG4_MUS 3
#define FEEDTIME_MUS 4
#define SONG6_MUS 5
#define BOSS_MUS 6
#define SONG8_MUS 7
#define MENU_MUS 8
#define HAPPYSNG_MUS 9
#define HIGHSCORE_MUS 10
#define LASTMUSIC 11
